mixregls.function <- function(obj, ...) {
	UseMethod("mixregls.function")
}

mixregls.function.default <- function(IDS, Y, X0, U0, W0, P=ncol(X0), R=ncol(U0), S=ncol(W0), PNINT=0, RNINT=0, SNINT=0, CONV=.0005, NQ=11, AQUAD=1, MAXIT=300, NCENT=0, NCOV=1, labels=list()) {
	#Need to do some error checking on input parameters
	if(PNINT != 1 && PNINT != 0)
		PNINT <- 0
	if(RNINT != 1 && RNINT != 0)
		RNINT <- 0
	if(SNINT != 1 && SNINT != 0)
		SNINT <- 0
	if(CONV <= 0 || CONV > .01)
		CONV <- .0005
	if(NQ <= 0 || NQ >= 100)
		NQ = 11
	if(AQUAD != 1 && AQUAD != 0)
		AQUAD <- 1
	if(MAXIT <= 0 && MAXIT > 10000)
		MAXIT <- 300
	if(NCENT != 1 && NCENT != 0)
		NCENT <- 0
	if(NCOV != 0 && NCOV != 1 && NCOV != 2)
		NCOV <- 1
	if(is.null(P) || P < 0)
		P <- 1
	if(is.null(R) || R < 0)
		R <- 1
	if(is.null(S) || S < 0)
		S <- 1
	NC <- length(unique(IDS))
	N  <- length(Y)

	Padj <- P+1-PNINT
	Radj <- R+1-RNINT
	Sadj <- S+1-SNINT

# call Fortran DLL
	mixregls=.Fortran("mixreglsr", matrix(as.integer(IDS),N,1), matrix(as.double(Y),N,1), matrix(as.double(X0),N,P),
				matrix(as.double(U0),N,R), matrix(as.double(W0),N,S), as.integer(N),as.integer(NC), as.integer(P), 
				as.integer(R), as.integer(S), as.integer(PNINT), as.integer(RNINT), as.integer(SNINT), as.double(CONV), 
				as.integer(NQ), as.integer(AQUAD), as.integer(MAXIT), as.integer(NCENT), as.integer(NCOV), 
				dev1=as.double(0.0),beta1=matrix(as.double(0.0),Padj,1), alpha1=matrix(as.double(0.0),Radj,1),
				tau1=matrix(as.double(0.0),1,1), se1=matrix(as.double(0.0),Padj+Radj+1,1),
				thet1=matrix(as.double(0.0),NC,1), pvar1=matrix(as.double(0.0),NC,1), zerr1=matrix(as.double(0.0),N,1), 
				dev2=as.double(0.0), beta2=matrix(as.double(0.0),Padj,1), alpha2=matrix(as.double(0.0),Radj,1), 
				tau2=matrix(as.double(0.0),Sadj,1), se2=matrix(as.double(0.0),Padj+Radj+Sadj,1), 
				thet2=matrix(as.double(0.0),NC,1), pvar2=matrix(as.double(0.0),NC,1), zerr2=matrix(as.double(0.0),N,1),
				dev3=as.double(0.0), beta3=matrix(as.double(0.0),Padj,1), alpha3=matrix(as.double(0.0),Radj,1), 
				tau3=matrix(as.double(0.0),Sadj,1), se3=matrix(as.double(0.0),Padj+Radj+Sadj+NCOV+1,1), 
				spar3=matrix(as.double(0.0),NCOV+1,1), thet3=matrix(as.double(0.0),NC,2),
				pvar3=matrix(as.double(0.0),NC,3), zerr3=matrix(as.double(0.0),N,1))
	result <- list(dev1=mixregls$dev1, alpha1=mixregls$alpha1, beta1=mixregls$beta1, tau1=mixregls$tau1, 
			se1=mixregls$se1, thet1=mixregls$thet1, pvar1=mixregls$pvar1, zerr1=mixregls$zerr1, 
			dev2=mixregls$dev2, alpha2=mixregls$alpha2, beta2=mixregls$beta2, tau2=mixregls$tau2,
			se2=mixregls$se2, thet2=mixregls$thet2, pvar2=mixregls$pvar2, zerr2=mixregls$zerr2, 
			dev3=mixregls$dev3, alpha3=mixregls$alpha3, beta3=mixregls$beta3, tau3=mixregls$tau3, se3=mixregls$se3,
			spar3=mixregls$spar3, thet3=mixregls$thet3, pvar3=mixregls$pvar3, zerr3=mixregls$zerr3, 
			P=P, R=R, S=S, PNINT=PNINT, RNINT=RNINT, SNINT=SNINT,	CONV=CONV, NQ=NQ, AQUAD=AQUAD, MAXIT=MAXIT, NCENT=NCENT, NCOV=NCOV, 
			labels=list(labelsY=labels$labelsY, labelsX0=labels$labelsX, labelsU=labels$labelsU, labelsW=labels$labelsW))
	class(result) <- "mixregls_output"
	return(result)
}

mixregls.function.mixregls_input <- function(obj, PNINT=obj$PNINT, RNINT=obj$RNINT, SNINT=obj$SNINT, CONV=obj$CONV, NQ=obj$NQ,
							   AQUAD=obj$AQUAD, MAXIT=obj$MAXIT, NCENT=obj$NCENT, NCOV=obj$NCOV, ...) {
	result <- mixregls.function.default(obj$IDS, obj$Y, obj$X0, obj$U0, obj$W0, obj$P, obj$R, obj$S,
							PNINT, RNINT, SNINT, CONV, NQ, AQUAD, MAXIT, NCENT, NCOV, labels=obj$labels)
	return(result);
}

print.mixregls_output <- function(x, ...) {
	Padj <- x$P-x$PNINT
	Radj <- x$R-x$RNINT
	Sadj <- x$S-x$SNINT
	PRadj <- Padj+Radj+3
	NCOVadj <- x$NCOV+1
	all3adj <- Padj+Radj+Sadj+3

	names = c(paste0("beta", 0:Padj), paste0("alpha", 0:Radj), paste0("tau", 0:Sadj), paste0("S", 1:NCOVadj))
	if(length(x$labels) > 0) {
		labelsX = x$labels$labelsX
		if(!x$PNINT)
			labelsX = c("Intercept", labelsX)
		labelsU = x$labels$labelsU
		if(!x$RNINT)
			labelsU = c("Intercept", labelsU)
		labelsW = x$labels$labelsW
		if(!x$SNINT)
			labelsW = c("Intercept", labelsW)
		names = c(paste("beta", labelsX), paste("alpha", labelsU), paste("tau", labelsW), paste0("S", 1:NCOVadj))
		cat("Dependent variable: ", x$labels$labelsY, "\n")
	}

	cat("Model 1:\n-2 ln L: ", x$dev1, "\n")
	values1 = c(x$beta1, x$alpha1, x$tau1)
	stderrs1 = x$se1
	zvals1 = values1/stderrs1
	pvals1 = sapply(abs(zvals1), dnorm)
	pvals1[pvals1 < .000005] = 0
	df1 = data.frame(cbind(values1, stderrs1, zvals1, pvals1), row.names=names[1:PRadj])
	names(df1) = c("Estimate", "AsymStdErr", "z-value", "p-value")
	print(df1, digits=4)

	cat("\nModel 2:\n-2 ln L: ", x$dev2, "\n")
	values2 = c(x$beta2, x$alpha2, x$tau2)
	stderrs2 = x$se2
	zvals2 = values2/stderrs2
	pvals2 = sapply(abs(zvals2), dnorm)
	pvals2[pvals2 < .000005] = 0
	df2 = data.frame(cbind(values2, stderrs2, zvals2, pvals2), row.names=names[1:all3adj])
	names(df2) = c("Estimate", "AsymStdErr", "z-value", "p-value")
	print(df2, digits=4)
	
	cat("\nModel 3:\n-2 ln L: ", x$dev3, "\n")
	values3 = c(x$beta3, x$alpha3, x$tau3, x$spar3)
	stderrs3 = x$se3
	zvals3 = values3/stderrs3
	pvals3 = sapply(abs(zvals3), dnorm)
	pvals3[pvals3 < .000005] = 0
	df3 = data.frame(cbind(values3, stderrs3, zvals3, pvals3), row.names=names)
	names(df3) = c("Estimate", "AsymStdErr", "z-value", "p-value")
	print(df3, digits=4)
}

mixregls.formula <- function(formula, id, data, PNINT = 0, RNINT = 0, SNINT = 0, CONV = 0.0005, NQ = 11, AQUAD = 1, MAXIT = 200, NCENT = 0, NCOV = 1) {
	library("Formula")
  	stopifnot(require("Formula"))
  	formula <- as.Formula(formula)

	termsX = terms(formula(formula, rhs=1))
	labelsX = labels(termsX)
	P = length(labelsX)
	PNINT = !attr(termsX, "intercept") && P > 0
	attr(termsX, "intercept") = 0
	X0 = model.matrix(termsX, data=data)

	termsU = terms(formula(formula, rhs=2))
	labelsU = labels(termsU)
	R = length(labelsU)
	RNINT = !attr(termsU, "intercept") && R > 0
	attr(termsU, "intercept") = 0
	U0 = model.matrix(termsU, data=data)

	termsW = terms(formula(formula, rhs=3))
	labelsW = labels(termsW)
	S = length(labelsW)
	SNINT = !attr(termsW, "intercept") && S > 0
	attr(termsW, "intercept") = 0
	W0 = model.matrix(termsW, data=data)

  	Y <- model.part(formula, data = data, lhs = NULL)
	labelsY = colnames(Y)
	tf1 = is.finite(rowSums(Y))
	Y = Y[tf1,]
	IDS = data[id][[1]][tf1]

	#filtering missing data
	tf = is.finite(rowSums(X0)) & is.finite(rowSums(U0)) & is.finite(rowSums(W0)) & is.finite(Y)
	X0 = as.matrix(X0[tf,])
	U0 = as.matrix(U0[tf,])
	W0 = as.matrix(W0[tf,])
	Y = as.matrix(Y[tf])
  	IDS <- IDS[tf]

	result <- list(IDS=IDS, Y=Y, X0=X0, U0=U0, W0=W0, P=P, R=R, S=S, PNINT=PNINT, RNINT=RNINT, SNINT=SNINT, CONV=CONV, 
			   NQ=NQ, AQUAD=AQUAD, MAXIT=MAXIT, NCENT=NCENT, NCOV=NCOV, labels=list(labelsY=labelsY, labelsX0=labelsX, labelsU=labelsU, labelsW=labelsW))
	class(result) <- "mixregls_input"
	return(result)
}

mixregls.combined <- function(formula, id, data, PNINT = 0, RNINT = 0, SNINT = 0, CONV = 0.0005, NQ = 11, AQUAD = 1, MAXIT = 200, NCENT = 0, NCOV = 1) {
	m <- mixregls.formula(formula, id, data, PNINT, RNINT, SNINT, CONV, NQ, AQUAD, MAXIT, NCENT, NCOV)
	return(mixregls.function(m))
}